from parameterized import parameterized
import unittest
from typing import List

import cedarpy

from shared import pretty_format, load_file_as_json, load_file_as_str


def custom_name_func(testcase_func, param_num, param):
    # print(f'{type(param.args)} {param.args}')
    return "%s_%s__%s" %(
        testcase_func.__name__,
        param_num,
        parameterized.to_safe_name("__".join(parameterized.to_safe_name(str(x)) for x in param.args)),
    )


def get_authz_test_params_for_test_suite(test_kind: str, test_suite: str) -> list:
    """Get authorization test params for a cedar-integration-tests test suite
    :param test_kind is one of the kinds of tests organized by directory in the cedar-integration-tests/tests
    directory, e.g. example_use_cases_doc
    :param test_suite is the test suite's 'id', which is the name of the file without the `.json`, e.g. '1a'
    """
    # Load the test data
    cedar_int_tests_base = "resources/cedar-integration-tests"
    test_def: dict = load_file_as_json(f"{cedar_int_tests_base}/tests/{test_kind}/{test_suite}.json")
    print(f'loading tests defined for use case: {test_suite}')

    policies_file_name: str = test_def['policies']
    entities_file_name: str = test_def['entities']
    schema_file_name: str = test_def['schema']
    should_validate: bool = test_def['should_validate']
    queries: List[dict] = test_def['queries']
    policies: str = load_file_as_str(f"{cedar_int_tests_base}/{policies_file_name}")
    entities: list = load_file_as_json(f"{cedar_int_tests_base}/{entities_file_name}")
    schema: object = load_file_as_json(f"{cedar_int_tests_base}/{schema_file_name}")

    testing_params = []

    for query in queries:
        testing_params.append((policies,
                               entities,
                               schema,
                               should_validate,
                               query))

    print(f'selected {len(testing_params)} test cases for {test_suite}:\n{pretty_format(testing_params)}')
    return testing_params


class BaseDataDrivenCedarIntegrationTestCase(unittest.TestCase):

    def exec_authz_query_with_assertions(self,
                                         policies: str,
                                         entities: list,
                                         schema: dict,
                                         should_validate: bool,  # ignored; currently don't have the equivalent
                                         query: dict) -> None:
        print(f"executing authz query:\n{pretty_format(query)}")
        request = {
            'principal': query['principal'],
            'action': query['action'],
            'resource': query['resource'],
            'context': query.get('context', {}),
        }
        authz_result: cedarpy.AuthzResult = cedarpy.is_authorized(request=request, policies=policies, entities=entities,
                                                                schema=schema)

        description = query['desc']
        self.assertEqual(query['decision'], authz_result.decision.value,
                         msg=f'unexpected decision for query desc: {description}')
        # 'reason' spelling is correct here, but a debatable choice as it's a list
        # 'reason' matches the (Rust) Decision enum but Java API has exposed as reasons (plural)
        self.assertEqual(query['reasons'], authz_result.diagnostics.reasons,
                         msg=f'unexpected errors for query desc: {description}')
        self.assertEqual(query['errors'], authz_result.diagnostics.errors,
                         msg=f'unexpected errors for query desc: {description}')


class CedarExampleUseCasesIntegrationTestCase(BaseDataDrivenCedarIntegrationTestCase):

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "1a"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_1a(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "2a"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_2a(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "2b"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_2b(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "2c"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_2c(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "3a"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_3a(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "3b"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_3b(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "3c"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_3c(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "4a"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_4a(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "4c"),
                          name_func=custom_name_func)
    @unittest.skip(reason="A couple of requests failing here; true reason TBD")
    def test_example_use_cases_doc_4c(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "4d"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_4d(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "4e"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_4e(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "4f"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_4f(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("example_use_cases_doc", "5b"),
                          name_func=custom_name_func)
    def test_example_use_cases_doc_5b(self,
                                      policies: str,
                                      entities: list,
                                      schema: dict,
                                      should_validate: bool,  # ignored; currently don't have the equivalent
                                      query: dict):

        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)


class CedarIPIntegrationTestCase(BaseDataDrivenCedarIntegrationTestCase):

    @parameterized.expand(get_authz_test_params_for_test_suite("ip", "1"),
                          name_func=custom_name_func)
    def test_ip_1(self,
                  policies: str,
                  entities: list,
                  schema: dict,
                  should_validate: bool,  # ignored; currently don't have the equivalent
                  query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("ip", "2"),
                          name_func=custom_name_func)
    def test_ip_2(self,
                  policies: str,
                  entities: list,
                  schema: dict,
                  should_validate: bool,  # ignored; currently don't have the equivalent
                  query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("ip", "3"),
                          name_func=custom_name_func)
    def test_ip_3(self,
                  policies: str,
                  entities: list,
                  schema: dict,
                  should_validate: bool,  # ignored; currently don't have the equivalent
                  query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)


class CedarMultiIntegrationTestCase(BaseDataDrivenCedarIntegrationTestCase):

    @parameterized.expand(get_authz_test_params_for_test_suite("multi", "1"),
                          name_func=custom_name_func)
    def test_multi_1(self,
                     policies: str,
                     entities: list,
                     schema: dict,
                     should_validate: bool,  # ignored; currently don't have the equivalent
                     query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("multi", "2"),
                          name_func=custom_name_func)
    def test_multi_2(self,
                     policies: str,
                     entities: list,
                     schema: dict,
                     should_validate: bool,  # ignored; currently don't have the equivalent
                     query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("multi", "3"),
                          name_func=custom_name_func)
    def test_multi_3(self,
                     policies: str,
                     entities: list,
                     schema: dict,
                     should_validate: bool,  # ignored; currently don't have the equivalent
                     query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("multi", "4"),
                          name_func=custom_name_func)
    @unittest.skip(reason="12 pass, 1 fails")
    def test_multi_4(self,
                     policies: str,
                     entities: list,
                     schema: dict,
                     should_validate: bool,  # ignored; currently don't have the equivalent
                     query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("multi", "5"),
                          name_func=custom_name_func)
    @unittest.skip(reason="Depends on unspecified principal, which is (currently) unsupported by is_authorized, i.e. principal is a required parameter")
    def test_multi_5(self,
                     policies: str,
                     entities: list,
                     schema: dict,
                     should_validate: bool,  # ignored; currently don't have the equivalent
                     query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)


class CedarDecimalIntegrationTestCase(BaseDataDrivenCedarIntegrationTestCase):

    @parameterized.expand(get_authz_test_params_for_test_suite("decimal", "1"),
                          name_func=custom_name_func)
    def test_decimal_1(self,
                       policies: str,
                       entities: list,
                       schema: dict,
                       should_validate: bool,  # ignored; currently don't have the equivalent
                       query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)

    @parameterized.expand(get_authz_test_params_for_test_suite("decimal", "2"),
                          name_func=custom_name_func)
    def test_decimal_2(self,
                       policies: str,
                       entities: list,
                       schema: dict,
                       should_validate: bool,  # ignored; currently don't have the equivalent
                       query: dict):
        self.exec_authz_query_with_assertions(policies=policies, entities=entities, schema=schema,
                                              should_validate=should_validate,
                                              query=query)




import json
import os
import pprint
from typing import Union


def pretty_format(o: object) -> str:
    """Pretty print an object representation"""
    return pprint.pformat(object=o, indent=2, width=120)


def load_file_as_json(relative_file_path: str, base_file=__file__) -> Union[object, list, dict]:
    path = construct_path_relative_to_current_module(relative_file_path, base_file)

    try:
        with open(path) as json_file:
            obj = json.load(json_file)
            return obj
    except FileNotFoundError:
        print("File could not be found at: {}".format(path))
        raise


def load_file_as_str(relative_file_path: str, base_file=__file__) -> str:
    path = construct_path_relative_to_current_module(relative_file_path, base_file)

    try:
        with open(path) as f:
            return f.read()
    except FileNotFoundError:
        print("File could not be found at: {}".format(path))
        raise


def construct_path_relative_to_current_module(relative_file_path, base_file=__file__):
    my_path = os.path.abspath(os.path.dirname(base_file))
    path = os.path.join(my_path, relative_file_path)
    return path



import datetime
from typing import Union


def load_file_as_json(relative_file_path: str) -> Union[object, list, dict]:
    import shared
    return shared.load_file_as_json(relative_file_path=relative_file_path,
                                    base_file=__file__)


def load_file_as_str(relative_file_path: str) -> str:
    import shared
    return shared.load_file_as_str(relative_file_path=relative_file_path,
                                   base_file=__file__)


def utc_now() -> datetime.datetime:
    """
    Get the current time in UTC.
    """
    return datetime.datetime.now(datetime.timezone.utc)



import unittest

from cedarpy import AuthzResult, Decision, Diagnostics


class AuthzResultTestCase(unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()

        self.allow_authz_resp = {
            "decision": "Allow",
            "diagnostics": {
                "reason": ["policy0"],
                "errors": []
            },
            "metrics": {"authz_duration_micros": 42}
        }

        self.deny_authz_resp = {
            'decision': 'Deny',
            'diagnostics': {
                'errors': ['while evaluating policy policy2, encountered the '
                           'following error: record does not have the '
                           'required attribute: authenticated'],
                'reason': []
            },
            "metrics": {"authz_duration_micros": 99}
        }

    def test_decision_property_when_Allow(self):
        authz_result = AuthzResult(self.allow_authz_resp)
        self.assertEqual(Decision.Allow, authz_result.decision)
        self.assertEqual(Decision.Allow, authz_result['decision'],
                         msg="decision should be available via subscript")
        self.assertTrue(authz_result.allowed)

    def test_decision_property_when_Deny(self):
        authz_result = AuthzResult(self.deny_authz_resp)
        self.assertEqual(authz_result.decision, Decision.Deny)
        self.assertEqual(Decision.Deny, authz_result['decision'])
        self.assertFalse(authz_result.allowed)

    def test_diagnostics_are_available(self):
        for authz_resp in [
            self.allow_authz_resp,
            self.deny_authz_resp,
        ]:
            authz_result = AuthzResult(authz_resp)
            self.assertIsNotNone(authz_result.diagnostics)

            self.assertEqual(authz_resp['diagnostics']['errors'],
                             authz_result.diagnostics.errors)

            self.assertEqual(authz_resp['diagnostics']['reason'],
                             authz_result.diagnostics.reasons,
                             f"expected 'reason' key (singular) to be mapped to reasons property (plural)"
                             f"; authz_resp: {authz_resp}")

    def test_metrics_are_available(self):
        missing_metrics = {}
        empty_metrics = {
            "metrics": {}  # empty metrics
        }
        for authz_resp in [
            self.allow_authz_resp,
            self.deny_authz_resp,
            empty_metrics,
            missing_metrics,
        ]:
            authz_result = AuthzResult(authz_resp)
            self.assertIsNotNone(authz_result.metrics,
                                 msg=f"metrics were none for {authz_resp}")

            if 'metrics' in authz_resp:
                self.assertEqual(authz_resp['metrics'],
                                 authz_result.metrics)
            else:
                self.assertEqual({}, authz_result.metrics)

    def test__getitem__makes_properties_subscriptable(self):
        for authz_resp in [
            self.allow_authz_resp,
            self.deny_authz_resp,
        ]:
            authz_result = AuthzResult(authz_resp)
            
            self.assertEqual(authz_result.decision, authz_result['decision'])
            self.assertEqual(authz_result.allowed, authz_result['allowed'])
            self.assertEqual(authz_result.diagnostics, authz_result['diagnostics'])
            self.assertEqual(authz_result.metrics, authz_result['metrics'])


class DiagnosticsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()

        self.allow_diagnostics_resp = {
            "errors": [],
            "reason": ["policy0"],
        }

        self.deny_diagnostics_resp = {
            'errors': ['while evaluating policy policy2, encountered the '
                       'following error: record does not have the '
                       'required attribute: authenticated'],
            'reason': [],
        }

    def test_errors_are_resolved(self):
        for diagnostics_resp in [self.allow_diagnostics_resp, self.deny_diagnostics_resp]:
            diagnostics = Diagnostics(diagnostics_resp)
            self.assertEqual(diagnostics_resp['errors'],
                             diagnostics.errors)

    def test_reasons_are_resolved(self):
        for diagnostics_resp in [self.allow_diagnostics_resp, self.deny_diagnostics_resp]:
            diagnostics = Diagnostics(diagnostics_resp)
            self.assertEqual(diagnostics_resp['reason'],
                             diagnostics.reasons)



import unittest

from textwrap import dedent

from cedarpy import format_policies

class FormatPolicyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_policy_gets_formatted(self):
        input_policy = dedent("""
            permit(
                principal,
                action == Action::"edit",
                resource
            )
            when {
                resource.owner == principal
            };
        """).strip()

        expect_result = dedent("""
            permit (
              principal,
              action == Action::"edit",
              resource
            )
            when { resource.owner == principal };
        """).strip()

        actual_result = format_policies(input_policy, indent_width=2)

        self.assertEqual(expect_result, actual_result)

    def test_policy_formatting_error(self):
        input_policy = dedent("""
            invalid(
                principal,
                action == Action::"edit",
                resource
            )
            when {
                resource.owner == principal
            };
        """).strip()

        try:
            format_policies(input_policy, indent_width=2)
            self.fail("should have failed to parse")
        except ValueError as e:
            pass



import json
from copy import copy
from enum import Enum
from typing import Union, List, Any

from cedarpy import _internal


def echo(s: str) -> str:
    return _internal.echo(s)


class Decision(Enum):
    Allow = 'Allow'
    Deny = 'Deny'
    NoDecision = 'NoDecision'


class Diagnostics:

    def __init__(self, diagnostics: dict) -> None:
        super().__init__()
        self._diagnostics: dict = diagnostics

    @property
    def errors(self) -> List[str]:
        return self._diagnostics.get('errors', list())

    @property
    def reasons(self) -> List[str]:
        # (intentionally) map 'reason' key in diagnostics dict to 'reasons' property (plural)
        return self._diagnostics.get('reason', list())


class AuthzResult:
    def __init__(self, authz_resp: dict) -> None:
        super().__init__()
        self._authz_resp = authz_resp
        self._diagnostics = Diagnostics(self._authz_resp.get('diagnostics', {}))

    @property
    def decision(self) -> Decision:
        return Decision[self._authz_resp['decision']]

    @property
    def allowed(self) -> bool:
        return Decision.Allow == self.decision

    @property
    def correlation_id(self) -> Decision:
        return self._authz_resp.get('correlation_id', None)

    @property
    def diagnostics(self) -> Diagnostics:
        return self._diagnostics

    @property
    def metrics(self) -> dict:
        return self._authz_resp.get('metrics', {})

    def __getitem__(self, __name: str) -> Any:
        return getattr(self, __name)


def is_authorized(request: dict,
                  policies: str,
                  entities: Union[str, List[dict]],
                  schema: Union[str, dict, None] = None,
                  verbose: bool = False) -> AuthzResult:
    """Evaluate whether the request is authorized given the parameters.

    :param request is a Cedar-style request object containing a principal, action, resource, and (optional) context;
    context may be a dict (preferred) or a string
    :param policies is a str containing all the policies in the Cedar PolicySet
    :param entities a list of entities or a json-formatted string containing the list of entities to
    include in the evaluation
    :param schema (optional) dictionary or json-formatted string containing the Cedar schema
    :param verbose (optional) boolean determining whether to enable verbose logging output within the library

    :returns an AuthzResult

    """
    return is_authorized_batch(requests=[request],
                               policies=policies,
                               entities=entities,
                               schema=schema,
                               verbose=verbose)[0]


def is_authorized_batch(requests: List[dict],
                        policies: str,
                        entities: Union[str, List[dict]],
                        schema: Union[str, dict, None] = None,
                        verbose: bool = False) -> List[AuthzResult]:
    """Evaluate whether a batch of requests are authorized given the other parameters.  Each request is evaluated
    independently and results in an AuthzResult per request.

    :param requests is list of Cedar-style request objects containing a principal, action, resource, and (optional) context;
    context may be a dict (preferred) or a string
    :param policies is a str containing all the policies in the Cedar PolicySet
    :param entities a list of entities or a json-formatted string containing the list of entities to
    include in the evaluation
    :param schema (optional) dictionary or json-formatted string containing the Cedar schema
    :param verbose (optional) boolean determining whether to enable verbose logging output within the library

    :returns a list of AuthzResults, in same order as the requests

    """
    requests_local = []
    for request in requests:
        if "context" in request:
            context = request["context"]
            if isinstance(context, dict):
                # ok user provided context as a dictionary, lets flatten it for them
                context_json_str = json.dumps(context)
                request = copy(request)
                request["context"] = context_json_str
            elif context is None:
                request = copy(request)
                del request["context"]

        requests_local.append(request)

    if isinstance(entities, str):
        pass
    elif isinstance(entities, list):
        entities = json.dumps(entities)

    if schema is not None:
        if isinstance(schema, str):
            pass
        elif isinstance(schema, dict):
            schema = json.dumps(schema)

    authz_result_strs: List[str] = _internal.is_authorized_batch(requests_local, policies, entities, schema, verbose)
    authz_result_objs: List[dict] = []

    for authz_result_str in authz_result_strs:
        authz_result_objs.append(json.loads(authz_result_str))
        
    authz_results: List[AuthzResult] = []
    for response_obj in authz_result_objs:
        authz_results.append(AuthzResult(response_obj))

    return authz_results


def format_policies(policies: str,
                    line_width: int = 80,
                    indent_width: int = 2) -> str:
    """Format the provided policies according to the Cedar conventions.

    :param policies is a str containing the policies to be formatted
    :param line_width (optional) is the desired maximum line length
    :param indent_width (optional) is the desired indentation width

    :returns the formatted policy
    :raises ValueError: if the input policies cannot be parsed
    """
    return _internal.format_policies(policies, line_width, indent_width)
